/*
 * Copyright (c) 2018 Nordic Semiconductor ASA
 *
 * SPDX-License-Identifier: LicenseRef-Nordic-5-Clause
 */
#include <zephyr/drivers/clock_control.h>
#include <zephyr/drivers/clock_control/nrf_clock_control.h>
#include <zephyr/drivers/gpio.h>
#include <zephyr/irq.h>
#include <zephyr/logging/log.h>
#include <nrf.h>
#include <esb.h>
#include <zephyr/device.h>
#include <zephyr/devicetree.h>
#include <zephyr/kernel.h>
#include <zephyr/types.h>
#include <zephyr/zephyr.h>
#include <cmsis_os.h>
#include <sys/printk.h>
// #include <sys/unistd.h>
// #include <zephyr/posix/unistd.h>

/* specify delay between greetings (in us); compute equivalent in ticks */
// #define TIMER_TICKS  1008
#define TIMER_TICKS 978
#define DELAY_TIMER_TICKS 250
#define TIMER_SEC_TICKS  1000000
#define ECG_FW_CH_LENTYPE 0x5A // 5A for MR400. 59 for Expression
#define ECG_FW_CH_PREAMBLE 0xAA
#define TOT_CMD_DATA_BYTES 4
#define TOT_SPARE_BYTES 3
#define TOT_ADDRESS_BYTES 4
#define TOT_ECG_DATA_BYTES 18
#define TOT_FWC_BYTES 26

#define NELEMS(x) (sizeof(x) / sizeof((x)[0]))

extern int dbg_events[];
extern int dbg_events_idx;
extern bool end_event;
extern bool tx_start_event;
void timer_callback(void const *arg);
void delay_function(void);
void delay_timer_callback(void const *arg);
osTimerDef(myTimer, timer_callback);
osTimerDef(myTimer_delay, delay_timer_callback);

//One second timer for debug print of total messages transmitted and received over past second
void sec_timer_callback(void const *arg);
osTimerDef(my_sec_timer, sec_timer_callback);

//Back channel message received for channel selection (commands 0x04 and 0x05)
unsigned char channel1_filter = 0x03;
unsigned char channel2_filter = 0x03;
int filter_recvd;

//Functions to set LED's 3 and 4 to visually indicate filter selection
int channel1_filter_led_set(void);
int channel2_filter_led_set(void);
void decimalToBinary(int num, int *val1, int *val2);

//Initialize ECG Packet for Forward Channel Transmission at 992 Hz
void initFwdChData(void);

//Channel 4 setting for Primary frequency of 2.440 GHz and Secondary frequency of 2.448 GHz
typedef enum hop_frequency
{
	primary_hop = 40,
	secondary_hop = 48
} hop_values;

int frequency_hop;
int network;
int networkGroup;
unsigned int networkAddress;
int counter;
bool packet_tx_done = true;
bool packet_ready = true;
bool delay_timer_flag = false;

//Counters for numer of TX messages and RX messages since last reset (1 second)
int				second_rx_cnt=0;
int				second_tx_cnt=0;

//ECG data sample variables for 2 Leads and 3 time slices (T0, T1, T2)
short int ch1DataA;
short int ch1DataB;
short int ch1DataC;
short int ch2DataA;
short int ch2DataB;
short int ch2DataC;

int rawLeadDataCh1;
int rawLeadDataCh2;
int led_count;

int led_value = 0;
osTimerId delay_timer_id;
bool tx_delay_done = false;
bool			print_metrics = false;

//ECG Waveform Simulation Array from MR400 divided by 10
static const short ECG_DAT[] = {
	-34, -32, -23, -23, -18, -18, -18, -12, -15, -20,
	-11, -18, -29, -32, -34, -29, -25, -28, -32, -19,
	-15, -6, -13, -16, -21, -21, -20, -14, -25, -28,
	-38, -36, -30, -28, -37, -25, -20, -13, -11, -9,
	-14, -18, -13, -17, -27, -28, -35, -39, -35, -31,
	-33, -36, -32, -25, -13, -17, -20, -13, -14, -15,
	-24, -25, -30, -29, -41, -35, -32, -39, -25, -19,
	-20, -16, -14, -18, -10, -7, -14, -15, -31, -27,
	-33, -29, -35, -23, -24, -19, -9, -8, -10, -12,
	-5, -7, -10, -9, -24, -31, -32, -33, -40, -36, // 100
	-29, -26, -17, -17, -11, -8, -12, -5, -3, -11,
	-13, -19, -29, -35, -32, -26, -26, -25, -18, -19,
	-15, -5, -6, -12, -14, -18, -18, -23, -32, -35,
	-25, -32, -28, -33, -27, -17, -14, -7, -14, -5,
	-10, -8, -18, -10, -18, -31, -27, -37, -31, -30,
	-28, -19, -6, -1, -1, -1, 6, 3, 10, 8,
	12, 1, 9, 6, 10, 14, 23, 18, 31, 36,
	46, 52, 49, 56, 56, 56, 50, 52, 49, 39,
	40, 46, 48, 51, 51, 55, 63, 65, 66, 73,
	69, 73, 68, 58, 57, 50, 46, 50, 43, 42, // 200
	52, 55, 50, 61, 57, 54, 60, 47, 45, 37,
	36, 19, 19, 20, 13, 16, 9, 11, 18, 16,
	20, 22, 17, 13, 4, -2, -14, -18, -26, -32,
	-29, -36, -25, -27, -22, -13, -12, -14, -15, -8,
	-9, -12, -10, -13, -32, -33, -33, -28, -25, -31,
	-28, -20, -19, -14, -10, -14, -6, -4, -13, -14,
	-20, -31, -37, -34, -28, -28, -28, -20, -21, -18,
	-19, -19, -12, -15, -8, -9, -27, -35, -38, -27,
	-34, -30, -34, -28, -17, -15, -7, -15, -6, -12,
	-10, -8, -13, -21, -22, -30, -28, -34, -32, -29, // 300
	-19, -18, -12, -12, -12, -17, -9, -15, -18, -25,
	-36, -29, -32, -39, -34, -37, -41, -39, -34, -35,
	-41, -43, -49, -49, -61, -68, -79, -82, -81, -81,
	-53, -41, -28, 6, 32, 62, 99, 123, 154, 187,
	217, 249, 277, 301, 331, 363, 391, 431, 452, 484,
	522, 540, 563, 572, 571, 557, 548, 539, 511, 490,
	452, 417, 382, 350, 313, 279, 240, 218, 176, 153,
	117, 88, 60, 28, 0, -34, -60, -79, -85, -96,
	-106, -110, -116, -103, -98, -99, -91, -80, -73, -72,
	-59, -54, -54, -46, -45, -45, -40, -38, -45, -43, // 400
	-36, -37, -32, -29, -34, -38, -34, -29, -30, -47,
	-44, -50, -47, -42, -43, -43, -37, -38, -35, -24,
	-37, -30, -32, -24, -24, -41, -49, -51, -41, -48,
	-44, -48, -43, -33, -31, -24, -31, -23, -28, -26,
	-23, -27, -35, -36, -44, -41, -36, -35, -45, -35,
	-21, -28, -15, -27, -20, -23, -17, -20, -28, -27,
	-32, -35, -30, -38, -28, -32, -31, -25, -15, -9,
	-12, -6, -7, -7, -15, -25, -28, -25, -24, -30,
	-15, -23, -23, -6, 2, 4, 5, -1, 7, 0,
	-5, -3, -17, -11, -15, -11, -18, -7, -10, -6, // 500
	1, 13, 8, 6, 13, 11, 9, 12, -1, -6,
	-5, -5, -11, -8, -2, 0, 7, 17, 20, 22,
	17, 25, 16, 21, 9, 5, 9, 3, 7, 12,
	0, 11, 17, 26, 28, 25, 35, 30, 40, 36,
	37, 21, 26, 25, 24, 17, 22, 17, 22, 31,
	44, 37, 41, 48, 42, 43, 46, 43, 36, 37,
	30, 33, 39, 41, 42, 51, 51, 56, 67, 65,
	71, 67, 74, 71, 65, 55, 52, 62, 56, 61,
	70, 64, 76, 79, 87, 92, 88, 94, 93, 94,
	88, 79, 78, 69, 72, 78, 80, 83, 82, 85, // 600
	91, 92, 104, 97, 93, 98, 82, 74, 74, 68,
	65, 70, 63, 61, 69, 71, 77, 75, 81, 77,
	71, 71, 70, 62, 51, 48, 37, 39, 44, 35,
	39, 39, 44, 41, 44, 34, 29, 25, 29, 12,
	13, -2, -10, -15, -23, -30, -19, -22, -29, -21,
	-21, -12, -15, -21, -22, -24, -33, -35, -53, -52,
	-52, -59, -55, -50, -48, -41, -42, -37, -34, -39,
	-31, -29, -36, -37, -53, -50, -55, -52, -58, -46,
	-47, -41, -43, -28, -31, -32, -37, -39, -30, -42,
	-45, -53, -55, -56, -51, -48, -53, -48, -51, -49, // 700
	-41, -49, -40, -45, -31, -41, -45, -52, -52, -49,
	-58, -53, -51, -49, -40, -39, -33, -33, -32, -25,
	-29, -22, -25, -33, -43, -48, -50, -45, -52, -43,
	-36, -35, -31, -22, -28, -19, -24, -24, -24, -30,
	-40, -42, -51, -49, -43, -53, -49, -50, -34, -39,
	-25, -25, -31, -23, -30, -34, -43, -43, -49, -52,
	-48, -43, -44, -48, -45, -26, -28, -21, -24, -29,
	-30, -31, -27, -39, -43, -42, -53, -47, -44, -51,
	-37, -32, -34, -31, -29, -22, -27, -23, -29, -29,
	-44, -51, -56, -39, -46, -47, -36, -32, -23, -23, // 800
	-26, -28, -21, -22, -25, -23, -37, -43, -44, -44,
	-50, -47, -40, -38, -30, -31, -25, -22, -27, -19,
	-17, -25, -26, -43, -40, -46, -43, -37, -37, -38,
	-31, -20, -18, -20, -21, -14, -16, -20, -19, -35,
	-43, -45, -46, -41, -37, -43, -38, -41, -27, -20,
	-28, -20, -25, -23, -20, -23, -30, -30, -38, -35,
	-42, -40, -38, -29, -28, -23, -24, -24, -17, -21,
	-15, -17, -24, -35, -39, -40, -47, -41, -44, -36,
	-24, -21, -12, -20, -23, -17, -17, -17, -23, -32,
	-33, -42, -39, -33, -43, -39, -28, -25, -18, -17, // 900
	-17, -23, -15, -22, -26, -23, -35, -41, -43, -38,
	-33, -35, -27, -24, -18, -20, -13, -16, -21, -21,
	-22, -18, -30, -34, -33, -32, -39, -36, -31, -30,
	-24, -15, -11, -9, -14, -18, -14, -20, -19, -34,
	-30, -35, -32, -39, -40, -29, -24, -16, -15, -6,
	-8, -14, -15, -17, -14, -28, -34, -35, -35, -42,
	-39, -33, -31, -23, -24, -19, -16, -20, -12, -9,
	-17, -18, -35, -32, -38, -35, -30, -30, -30, -24,
	-25, -22, -12, -13, -18, -20, -12, -12, -29, -37,
	-39, -28, -32768};

int total_items_ecg_dat = NELEMS(ECG_DAT);

//Structure used to set up Nordic Radio Packet and settings
typedef struct NetTable_t
{
	unsigned int addr;
	char cfgHi;
	char cfgLo1;
	char cfgLo2;
} NetTable;

//Table of Address Bytes, CFGHi, CFGLo1, and CFGLo2 for Channels 1-10
NetTable NetworkTable[] =
	{
		{0xBB44BBB5, 0xAF, 138, 154},
		{0x9B449B95, 0xAF, 72, 88},
		{0xBA44BAB5, 0xAF, 74, 90},
		{0x99449996, 0xAF, 80, 96},
		{0xA944A9A5, 0xAF, 70, 86},
		{0xA6565656, 0xAF, 144, 160},
		{0xC5556666, 0xAF, 110, 126},
		{0xA6669999, 0xAF, 108, 124},
		{0x96969696, 0xAF, 116, 132},
		{0xCCCC5555, 0xAF, 106, 122},
		{0xBBBBBBBB, 0xAF, 4, 20},
		{0x9B9B9B9B, 0xAF, 6, 22},
		{0xBABABABA, 0xAF, 12, 28},
		{0x99999999, 0xAF, 14, 30},
		{0xA9A9A9A9, 0xAF, 36, 52}
	};

// Structure for ECG Forward Channel Packet
typedef struct
{
	//	char 	wPreamble;
	//	char 	wAddress[TOT_ADDRESS_BYTES];
	char wLenType;
	char wECGData[TOT_ECG_DATA_BYTES];
	char wEcgSeqNumber;
	char wEcgStatusData;
} EcgForwardChannelPacket;

static EcgForwardChannelPacket EcgFWCPckt = {
	//		 0x00,
	//	    {0x00,0x00,0x00,0x00},
	0x00,
	{0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	 0x00, 0x00, 0x00, 0x00, 0x00, 0x00},
	0x00,
	0x00};

LOG_MODULE_REGISTER(esb_ptx, CONFIG_ESB_PTX_APP_LOG_LEVEL);
#if 0
static const struct gpio_dt_spec leds[] = {
	GPIO_DT_SPEC_GET(DT_ALIAS(led0), gpios),
	GPIO_DT_SPEC_GET(DT_ALIAS(led1), gpios),
	GPIO_DT_SPEC_GET(DT_ALIAS(led2), gpios),
	GPIO_DT_SPEC_GET(DT_ALIAS(led3), gpios),
};

BUILD_ASSERT(DT_SAME_NODE(DT_GPIO_CTLR(DT_ALIAS(led0), gpios),
			  DT_GPIO_CTLR(DT_ALIAS(led1), gpios)) &&
	     DT_SAME_NODE(DT_GPIO_CTLR(DT_ALIAS(led0), gpios),
			  DT_GPIO_CTLR(DT_ALIAS(led2), gpios)) &&
	     DT_SAME_NODE(DT_GPIO_CTLR(DT_ALIAS(led0), gpios),
			  DT_GPIO_CTLR(DT_ALIAS(led3), gpios)),
	     "All LEDs must be on the same port");
#endif

/* The devicetree node identifier for the "led3" alias. */
#define LED0_NODE DT_ALIAS(led0)
static const struct gpio_dt_spec led = GPIO_DT_SPEC_GET(LED0_NODE, gpios);
#define LED1_NODE DT_ALIAS(led1)
static const struct gpio_dt_spec led_1 = GPIO_DT_SPEC_GET(LED1_NODE, gpios);
#define LED2_NODE DT_ALIAS(led2)
static const struct gpio_dt_spec led_3 = GPIO_DT_SPEC_GET(LED2_NODE, gpios);
#define LED3_NODE DT_ALIAS(led3)
static const struct gpio_dt_spec led_4 = GPIO_DT_SPEC_GET(LED3_NODE, gpios);
int led_count;

static bool ready = false;
static struct esb_payload rx_payload;

static struct esb_payload tx_payload = ESB_CREATE_PAYLOAD(0,
														  0x00,
														  0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
														  0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
														  0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
														  0x00,
														  0x00);

#define _RADIO_SHORTS_COMMON                                       \
	(RADIO_SHORTS_READY_START_Msk | RADIO_SHORTS_END_DISABLE_Msk | \
	 RADIO_SHORTS_ADDRESS_RSSISTART_Msk |                          \
	 RADIO_SHORTS_DISABLED_RSSISTOP_Msk)

//Function to Process Nordic Events
void event_handler(struct esb_evt const *event)
{
	ready = true;

	switch (event->evt_id)
	{
	case ESB_EVENT_TX_SUCCESS:
		// LOG_DBG("TX SUCCESS EVENT");
		break;
	case ESB_EVENT_TX_FAILED:
		LOG_DBG("TX FAILED EVENT");
		break;
	case ESB_EVENT_RX_RECEIVED:
		if (esb_read_rx_payload(&rx_payload) == 0)
		{
#if 0
			LOG_DBG("Packet received, len %d, data: "
				"0x%02x, 0x%02x, 0x%02x, 0x%02x, "
				"0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x ",
				rx_payload.length, rx_payload.data[0],
				rx_payload.data[1], rx_payload.data[2],
				rx_payload.data[3], rx_payload.data[4],
				rx_payload.data[5], rx_payload.data[6],
				rx_payload.data[7],rx_payload.data[8]);
#endif
#if 0
				if(rx_payload.data[1] == 0x04)
				{
					printk("Filter data 0x04 : 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x \r\n",rx_payload.data[1], rx_payload.data[2],
				rx_payload.data[3], rx_payload.data[4], rx_payload.data[5]);
					channel1_filter = rx_payload.data[2];
					//filter_recvd = 1;

				}
#endif
			//if ((rx_payload.data[1] == 0x05) && (channel2_filter != rx_payload.data[2]))
			if (rx_payload.data[1] == 0x05)
			{

				printk("Filter data 0x05 : 0x%02x, 0x%02x, 0x%02x, 0x%02x, 0x%02x \r\n", rx_payload.data[1], rx_payload.data[2],
					   rx_payload.data[3], rx_payload.data[4], rx_payload.data[5]);
				
				if (channel2_filter != rx_payload.data[2])
				{
					channel2_filter = rx_payload.data[2];
					filter_recvd = 2;
				}
			}
			second_rx_cnt++;	

		}
		break;
	}
}

//Function to start system clocks
int clocks_start(void)
{
	int err;
	int res;
	struct onoff_manager *clk_mgr;
	struct onoff_client clk_cli;

	clk_mgr = z_nrf_clock_control_get_onoff(CLOCK_CONTROL_NRF_SUBSYS_HF);
	if (!clk_mgr)
	{
		LOG_ERR("Unable to get the Clock manager");
		return -ENXIO;
	}

	sys_notify_init_spinwait(&clk_cli.notify);

	err = onoff_request(clk_mgr, &clk_cli);
	if (err < 0)
	{
		LOG_ERR("Clock request failed: %d", err);
		return err;
	}

	do
	{
		err = sys_notify_fetch_result(&clk_cli.notify, &res);
		if (!err && res)
		{
			LOG_ERR("Clock could not be started: %d", res);
			return res;
		}
	} while (err);

	LOG_DBG("HF clock started");
	return 0;
}

//Function to initialize Nordic Radio at Start Up for RX
int esb_initialize(void)
{
	int err;
	 
	//Nordic radio can have 8 Address Bytes.  MR400 implementation uses 4 Address Bytes in base_addr_0
	uint8_t base_addr_0[4] = {0x99, 0x44, 0x99, 0x96};
	uint8_t base_addr_1[4] = {0x00, 0x00, 0x00, 0x00};

	//Nordic address prefix MR400 uses 1 and it is set to LS Byte of Address Bytes
	uint8_t addr_prefix[8] = {0x96, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00};

	struct esb_config config = ESB_DEFAULT_CONFIG;

	config.protocol = ESB_PROTOCOL_ESB;
	config.retransmit_delay = 600;
	config.bitrate = ESB_BITRATE_1MBPS;
	config.event_handler = event_handler;
	config.mode = ESB_MODE_PRX;
	config.selective_auto_ack = true;
	config.payload_length = 9;

	err = esb_init(&config);

	if (err)
	{
		printk("esb_init failed \r\n");
		return err;
	}

	err = esb_set_base_address_0(base_addr_0);
	if (err)
	{
		return err;
	}

	err = esb_set_base_address_1(base_addr_1);
	if (err)
	{
		return err;
	}

	err = esb_set_prefixes(addr_prefix, ARRAY_SIZE(addr_prefix));
	if (err)
	{
		return err;
	}
	tx_payload.noack = 1;
	rx_payload.noack = 1;

	return 0;
}

//Function to set Nordic Radio in TX mode
int esb_initialize_tx(void)
{
	int err;


	struct esb_config config = ESB_DEFAULT_CONFIG;

	config.protocol = ESB_PROTOCOL_ESB;
	config.retransmit_delay = 600;
	config.bitrate = ESB_BITRATE_1MBPS;
	config.event_handler = event_handler;
	config.mode = ESB_MODE_PTX;
	config.selective_auto_ack = true;
	config.payload_length = 21;

	err = esb_init_tx(&config);

	if (err)
	{
		return err;
	}

	return 0;
}

//Function to set Nordic Radio in RX mode
int esb_initialize_rx(void)
{
	int err;


	struct esb_config config = ESB_DEFAULT_CONFIG;

	config.protocol = ESB_PROTOCOL_ESB;
	config.retransmit_delay = 600;
	config.bitrate = ESB_BITRATE_1MBPS;
	config.event_handler = event_handler;
	config.mode = ESB_MODE_PRX;
	config.selective_auto_ack = true;
	config.payload_length = 9;

	err = esb_init_rx(&config);
	if (err)
	{
		return err;
	}

	return 0;
}

//Function to initialize LED's 1, 3, 4
static int leds_init(void)
{
	int ret = 0;

	if (!device_is_ready(led.port))
	{
		printk(" error device_is_ready \r\n");
		return -ENODEV;
	}

	ret = gpio_pin_configure_dt(&led, GPIO_OUTPUT_ACTIVE);
	if (ret < 0)
	{
		printk(" error gpio_pin_configure_dt \r\n");
		return -2;
	}

	if (!device_is_ready(led_1.port))
	{
		printk(" error device_is_ready \r\n");
		return -ENODEV;
	}

	ret = gpio_pin_configure_dt(&led_1, GPIO_OUTPUT_ACTIVE);
	if (ret < 0)
	{
		printk(" error gpio_pin_configure_dt \r\n");
		return -2;
	}

	if (!device_is_ready(led_3.port))
	{
		printk(" error device_is_ready \r\n");
		return -ENODEV;
	}

	ret = gpio_pin_configure_dt(&led_3, GPIO_OUTPUT_ACTIVE);
	if (ret < 0)
	{
		printk(" error gpio_pin_configure_dt \r\n");
		return -2;
	}

	if (!device_is_ready(led_4.port))
	{
		printk(" error device_is_ready \r\n");
		return -ENODEV;
	}

	ret = gpio_pin_configure_dt(&led_4, GPIO_OUTPUT_ACTIVE);
	if (ret < 0)
	{
		printk(" error gpio_pin_configure_dt \r\n");
		return -2;
	}

	return 0;
}

//Function to initialize the Forward Channel Message.
void initFwdChData(void)
{
	int ret = 0;
	// Each FWC packet consists of 28 bytes (26 data bytes + 2 CRC bytes) which is transmitted at a rate of 992 Hz (1.008065 ms)
	// 1. Header - Preamble (8 Bits) + Address (32 Bits) + Length (4 Bits) + Type (4 Bits)
	// 2. ECG DATA "A" - Data1 (16 bits) + Data2 (16 Bits) + Data3 (16 bits)
	// 3. ECG DATA "B" - Data1 (16 bits) + Data2 (16 Bits) + Data3 (16 bits)
	// 4. ECG DATA "C" - Data1 (16 bits) + Data2 (16 Bits) + Data3 (16 bits)
	// 5. Status Information - Status Code (8 bits) + Status Data (8 bits)
	// 6. CRC - CRC data (16 bits). The CRC bytes are calculated by the Radio firmware block and not calculated here.

	// EcgFWCPckt.wPreamble = ECG_FW_CH_PREAMBLE; // Preamble
	EcgFWCPckt.wLenType = ECG_FW_CH_LENTYPE; // Length/type
	EcgFWCPckt.wEcgSeqNumber = 0x00;
	EcgFWCPckt.wEcgStatusData = 0x00;

	network = 4;
	// Get the Network Address from the network table
	networkAddress = NetworkTable[network - 1].addr;

	// EcgFWCPckt.wAddress[0] = (networkAddress >> 24) & 0xFF;
	// EcgFWCPckt.wAddress[1] = (networkAddress >> 16) & 0xFF;
	// EcgFWCPckt.wAddress[2] = (networkAddress >> 8) & 0xFF;
	// EcgFWCPckt.wAddress[3] = networkAddress & 0xFF;

	ch1DataA = 0;
	ch1DataB = 0;
	ch1DataC = 0;
	ch2DataA = 0;
	ch2DataB = 0;
	ch2DataC = 0;

	// ECG DATA "A"
	// DATA 1
	EcgFWCPckt.wECGData[0] = ch1DataA & 0xFF;
	EcgFWCPckt.wECGData[1] = (ch1DataA & 0xFF00) >> 8;

	// DATA 2
	EcgFWCPckt.wECGData[2] = ch2DataA & 0xFF;
	EcgFWCPckt.wECGData[3] = (ch2DataA & 0xFF00) >> 8;

	// DATA 3
	// The 0x55 value for Data A-3 prevents a long series of 0 bit values
	// in the packet transmission
	EcgFWCPckt.wECGData[4] = 0x55;
	EcgFWCPckt.wECGData[5] = 0x55;

	// ECG DATA "B"
	// DATA 1
	EcgFWCPckt.wECGData[6] = ch1DataB & 0xFF;
	EcgFWCPckt.wECGData[7] = (ch1DataB & 0xFF00) >> 8;

	// DATA 2
	EcgFWCPckt.wECGData[8] = ch2DataB & 0xFF;
	EcgFWCPckt.wECGData[9] = (ch2DataB & 0xFF00) >> 8;

	// DATA 3
	// The 0x55 value for Data B-3 prevents a long series of 0 bit values
	// in the packet transmission
	EcgFWCPckt.wECGData[10] = 0x55;
	EcgFWCPckt.wECGData[11] = 0x55;

	// ECG DATA "C"
	// DATA 1
	EcgFWCPckt.wECGData[12] = ch1DataC & 0xFF;
	EcgFWCPckt.wECGData[13] = (ch1DataC & 0xFF00) >> 8;

	// DATA 2
	EcgFWCPckt.wECGData[14] = ch2DataC & 0xFF;
	EcgFWCPckt.wECGData[15] = (ch2DataC & 0xFF00) >> 8;

	// DATA 3
	// The 0x55 value for Data C-3 prevents a long series of 0 bit values
	// in the packet transmission
	EcgFWCPckt.wECGData[16] = 0x55;
	EcgFWCPckt.wECGData[17] = 0x55;

	// Status Code/ Sequence #
	EcgFWCPckt.wEcgSeqNumber++;
	EcgFWCPckt.wEcgSeqNumber = EcgFWCPckt.wEcgSeqNumber & 0xFF;
	bool packet_tx_done = true;
	bool packet_ready = false;

	channel2_filter = 0x03;
	channel1_filter = 0x03;

	// Set the radio frequency and start rx
	frequency_hop = primary_hop;
	esb_set_rf_channel(frequency_hop);
	esb_start_rx();
	frequency_hop = secondary_hop;

	ret = gpio_pin_set(led_3.port, led_3.pin, 0);
	if (ret)
	{
		LOG_ERR("LED3 set 1 error: %d", ret);
	}
	ret = gpio_pin_set(led_4.port, led_4.pin, 1);
	if (ret)
	{
		LOG_ERR("LED4 set 1 error: %d", ret);
	}
}

void delay_function(void)
{
	delay_timer_flag = true;
	while (delay_timer_flag == false);
}

void delay_timer_callback(void const *arg)
{
	tx_delay_done = true;
}

void sec_timer_callback(void const *arg)
{
	print_metrics = true;
}

void decimalToBinary(int num, int *val1, int *val2) {
	// Stores binary representation of number.
	int binaryNum[3] = {0, 0, 0}; // Assuming 32 bit integer.
	int i = 0;

	for (; num > 0;)
	{
		binaryNum[i++] = num % 2;
		num /= 2;
	}

	*val1 = binaryNum[1];
	*val2 = binaryNum[0];
}

int channel1_filter_led_set(void)
{

	int ret = 0, value_led2 = 0, value_led4 = 0;

	decimalToBinary((channel1_filter - 2), &value_led2, &value_led4);

	ret = gpio_pin_set(led_3.port, led_3.pin, value_led2);
	if (ret)
	{
		LOG_ERR("LED3 set 1 error: %d", ret);
	}

	ret = gpio_pin_set(led_4.port, led_4.pin, value_led4);
	if (ret)
	{
		LOG_ERR("LED4 set 1 error: %d", ret);
	}
}

int channel2_filter_led_set(void)
{

	int ret = 0, value_led2 = 0, value_led4 = 0;
	decimalToBinary((channel2_filter - 2), &value_led2, &value_led4);

	ret = gpio_pin_set(led_3.port, led_3.pin, value_led2);
	if (ret)
	{
		LOG_ERR("LED3 set 1 error: %d", ret);
	}

	ret = gpio_pin_set(led_4.port, led_4.pin, value_led4);
	if (ret)
	{
		LOG_ERR("LED4 set 1 error: %d", ret);
	}
	filter_recvd = 0;
}

void timer_callback(void const *arg)
{

	int ret = 0, err = 0;

	//*************** Tasks to do here ****************
	//	1. In init(), set radio for RX
	//	2. Set radio Frequency Register with frequency variable
	//	3. Set Radio for TX (TX_EN) (There may be a sequence required and a wait for READY event)
	//*************************************************

	if ((packet_ready == true) && (packet_tx_done == false))
	{
		esb_stop_rx();

		// delay_function();
		esb_set_rf_channel(frequency_hop);

		esb_initialize_tx();

		// delay_function();

		esb_flush_tx();


		err = esb_write_payload(&tx_payload);
		if (err)
		{
			LOG_ERR("Payload write failed, err %d", err);
		}

		second_tx_cnt++;

		esb_flush_tx();
		// delay_function();
		// turn off/on led
		// usleep(10);
		// k_sleep(K_USEC(10));
		// printk("\r\n Inside timer handler, after write payload \r\n");
		printk("\n  \n");
		//if (esb_is_idle() != true)
		//{
		//	printk("ESB not idle \r\n");
		//}
		//printk("\r     \r");
		//printk("\r\r\r\r\r\r\r\r\r\r\n");
		packet_tx_done = true;
		//printk("BEfore while \r\n");
		//while((tx_start_event == true)&&(end_event == false))
		//{
			
			
		//}
		//printk("tx complete\r\n");
		//tx_start_event = false;
		//end_event = false;
		//esb_stop_tx();

#if 0
	led_value = !led_value;
	printk("led_value : %d \r\n",led_value);
	ret = gpio_pin_set(&led_1.port, led_1.pin, led_value);
	if (ret) {
		LOG_ERR("LED1 set 0 error: %d", ret);
	}

#endif
		ret = gpio_pin_toggle_dt(&led_1);
		if (ret < 0)
		{
			return;
		}

		esb_set_rf_channel(primary_hop);
		esb_initialize_rx();
		// delay_function();
		esb_start_rx();

	}
	//**************** After TX Success ***************
	//	4. Set Radio for RX (RX_EN) (There may be  a sequence required and a wait for READY event)
	//*************************************************
}

void main(void)
{
	int err, i = 0, ret = 0;

	osTimerId timer_id;
	osTimerId timer_sec_id;

	LOG_INF("Enhanced ShockBurst ptx sample");
	printk(" \r\n ** total_items_ecg_dat: %d **\r\n", total_items_ecg_dat);

	err = clocks_start();
	if (err)
	{
		return;
	}

	err = leds_init();
	if (err)
	{
		return;
	}

	err = esb_initialize();
	if (err)
	{
		LOG_ERR("ESB initialization failed, err %d", err);
		return;
	}

	initFwdChData();

	LOG_INF("Initialization complete");
	LOG_INF("Sending test packet");
	delay_timer_id = osTimerCreate_US(osTimer(myTimer_delay), osTimerOnce, NULL);
	timer_id = osTimerCreate_US(osTimer(myTimer), osTimerPeriodic, NULL);
	osTimerStart_US(timer_id, TIMER_TICKS);

	timer_sec_id = osTimerCreate_US(osTimer(my_sec_timer), osTimerPeriodic, NULL);
	osTimerStart_US(timer_sec_id, TIMER_SEC_TICKS);
	while (1)
	{
		if (packet_tx_done == true)
		{
			packet_ready == false;
			//********************* Tasks to do *************
			//	1. Set a frequency variable to Primary or Secondary frequency for next transmission based on hopping pattern/network
			//************************************************

			if (frequency_hop == primary_hop)
			{
				frequency_hop = secondary_hop;
			}
			else if (frequency_hop == secondary_hop)
			{
				frequency_hop = primary_hop;
			}
			else
			{
				frequency_hop = primary_hop;
			}

			// At Max QRS value, set LED 4 ON
			if (ECG_DAT[counter] == -572)
			{
				// turn on led
				ret = gpio_pin_set(&led, led.pin, 1);
				if (ret)
				{
					LOG_ERR("LED3 set 1 error: %d", ret);
				}
				led_count = 1;
			}

			// Move T-1 Data to T-2 Data and T Data to T-1 Data
			ch1DataC = ch1DataB;
			ch2DataC = ch2DataB;
			ch1DataB = ch1DataA;
			ch2DataB = ch2DataA;

			ch1DataA = ECG_DAT[counter];
			ch2DataA = ECG_DAT[counter];

			// ECG DATA "A"
			// DATA 1
			EcgFWCPckt.wECGData[0] = ch1DataA & 0xFF;
			EcgFWCPckt.wECGData[1] = (ch1DataA & 0xFF00) >> 8;

			// DATA 2
			EcgFWCPckt.wECGData[2] = ch2DataA & 0xFF;
			EcgFWCPckt.wECGData[3] = (ch2DataA & 0xFF00) >> 8;

			// DATA 3
			// The 0x55 value for Data A-3 prevents a long series of 0 bit values
			// in the packet transmission
			EcgFWCPckt.wECGData[4] = 0x55;
			EcgFWCPckt.wECGData[5] = 0x55;

			// ECG DATA "B"
			// DATA 1
			EcgFWCPckt.wECGData[6] = ch1DataB & 0xFF;
			EcgFWCPckt.wECGData[7] = (ch1DataB & 0xFF00) >> 8;

			// DATA 2
			EcgFWCPckt.wECGData[8] = ch2DataB & 0xFF;
			EcgFWCPckt.wECGData[9] = (ch2DataB & 0xFF00) >> 8;

			// DATA 3
			// The 0x55 value for Data B-3 prevents a long series of 0 bit values
			// in the packet transmission
			EcgFWCPckt.wECGData[10] = 0x55;
			EcgFWCPckt.wECGData[11] = 0x55;

			// ECG DATA "C"
			// DATA 1
			EcgFWCPckt.wECGData[12] = ch1DataC & 0xFF;
			EcgFWCPckt.wECGData[13] = (ch1DataC & 0xFF00) >> 8;

			// DATA 2
			EcgFWCPckt.wECGData[14] = ch2DataC & 0xFF;
			EcgFWCPckt.wECGData[15] = (ch2DataC & 0xFF00) >> 8;

			// DATA 3
			// The 0x55 value for Data C-3 prevents a long series of 0 bit values
			// in the packet transmission
			EcgFWCPckt.wECGData[16] = 0x55;
			EcgFWCPckt.wECGData[17] = 0x55;

			// Set sequence number for next packet transmission
			EcgFWCPckt.wEcgSeqNumber++;
			EcgFWCPckt.wEcgSeqNumber = EcgFWCPckt.wEcgSeqNumber & 0xFF;

			//*************** Tasks to do ******************
			//	1. Write EcgFWCPckt.wEcgStatusData = New Structure[EcgFWCPckt.wEcgSeqNumber & 0x3F]
			//**********************************************

			// Copy ECG packet data into transmit payload
			memcpy(tx_payload.data, &EcgFWCPckt, 21);

			// Set counter index for ECG simulated data
			if (counter < total_items_ecg_dat)
			{
				counter = counter + 1;
			}
			else
			{
				counter = 0;
			}

			if (led_count > 0)
			{
				led_count++;
			}
			if (led_count > 250)
			{
				// turn off led
				ret = gpio_pin_set(&led, led.pin, 0);
				if (ret)
				{
					LOG_ERR("LED3 set 0 error: %d", ret);
				}
				led_count = 0;
			}

			esb_flush_tx();

			packet_ready == true;
			packet_tx_done = false;
		}
		else
		{ // packet_tx_done = false

#if 0
			if(filter_recvd == 1)
			{	
				channel1_filter_led_set();
			}
#endif
			if (filter_recvd == 2)
			{
				channel2_filter_led_set();
			}

			if (delay_timer_flag == true)
			{
				osTimerStart_US(delay_timer_id, DELAY_TIMER_TICKS);
				while (tx_delay_done == false);
				tx_delay_done = false;
				osTimerStop(delay_timer_id);
				delay_timer_flag = false;
			}

			if( print_metrics == true )
			{
				printk("second_tx_cnt: %d     second_rx_cnt: %d \r\n", second_tx_cnt, second_rx_cnt);
				second_rx_cnt = 0;
				second_tx_cnt = 0;
				print_metrics = false;
			}
			
			//*************** Tasks to do ***********************
			//	1. Set up a structure of 64 chars to hold status data
			//	2. In init() function initialize structure with current value sent as data.
			//	3. Based on index (SequenceNumber & 0x3F), fill in designated RX data
			//		a. Plan is to fill in Filter setting with data received from MR400
			//		b. Based on Filter Setting, set battery remaining time to designated value
			//	4. Process any received data into Status Data
			//***************************************************
		}
		// k_sleep(K_MSEC(100));
	}
}
